#!/usr/bin/perl

use strict;

my ($use_68k, $use_z80, $use_6502, $use_020, # cpus
  $use_2151,$use_2203,$use_2413,$use_2610,$use_3812,$use_adpcm,$use_ay8910,
  $use_dac, $use_dxsmp, $use_ensoniq, $use_m6585, $use_msm5205, $use_namco,
  $use_qsound,$use_smp16bit,$use_toaplan2,$use_x1_010,$use_ymf278b,
  $use_ymz280b) = undef;
my @obj = ();

while ($_ = shift @ARGV) {
  my @drivers = ();
  open(F,"<$_") || die "open $_\n";
  my $file = $_;
  my $name = uc($_);
  $name =~ s/\..*//; # strip extension

  while (<F>) {
    next if (/\\\r?$/); # macro definition
    if (/^(GAME|CLONE|CLONE_ROMSW|CLONE_LOAD|CLONE_DSW|GME|GME_ROMSW|cps1_gameb?|cps1_game_270b?|cps1_cloneb?|cps1b4_game|cps1b4_clone|sfzch_game|cps1_clone_270b?|forgottn_game|forgottn_clone|cps2_game|cps2_region|cps2_clone|cps2_clone_no_parent|cps2_game_270|qsound_game|qsound_clone|sf2_game|sf2_clone)\( ?([a-z0-9A-Z_]+)/) {
      push @drivers,$2;
    } elsif (/(AddRead|AddWrite)/) {
      $use_68k = 1
    } elsif (/R24\[/) {
      $use_020 = 1;
    } elsif (/f3system.h/) {
      $use_020 = 1;
      $use_ensoniq = 1;
      $use_68k = 1;
    } elsif (/AddZ80/) {
      $use_z80 = 1;
    } elsif (/AddM6502/) {
      $use_6502 = 1;
    } elsif (/(taitosnd.h|tchnosnd.h)/) {
      $use_2610 = 1;
      $use_2151 = 1;
      $use_2203 = 1;
      $use_adpcm = 1;
    } elsif (/2151intf.h/) {
      $use_2151 = 1;
    } elsif (/2203intf.h/) {
      $use_2203 = 1;
    } elsif (/2413intf.h/) {
      $use_2413 = 1;
    } elsif (/2610intf.h/) {
      $use_2610 = 1;
    } elsif (/(3812intf.h|nichisnd.h)/) {
      $use_3812 = 1;
    } elsif (/adpcm.h/) {
      $use_adpcm = 1;
    } elsif (/ay8910.h/) {
      $use_ay8910 = 1;
    } elsif (/dac.h/) {
      $use_dac = 1;
    } elsif (/dxsmp.h/) {
      $use_dxsmp = 1;
    } elsif( /es5506.h/) {
      $use_ensoniq = 1;
    } elsif (/m6585.h/) {
      $use_m6585 = 1;
    } elsif (/msm5205.h/) {
      $use_msm5205 = 1;
    } elsif (/namco.h/) {
      $use_namco = 1;
    } elsif (/qsound.h/) {
      $use_qsound = 1;
    } elsif (/smp16bit.h/) {
      $use_smp16bit = 1;
    } elsif (/ymz280b.h/) {
      $use_ymz280b = 1;
    }
  } # while
  close(F);
  next if (!@drivers);
#   open(F,">>conf.h") || die "can't write to conf.h\n";
#   print F "#define DRIVER_$name\n";
#   close(F);
  open(F,">>drivers.h") || die "can't write to drivers.h\n";
  # print F "#ifdef DRIVER_$name\n";
  print F "// $file\n";
  foreach (@drivers) {
    print F "  DRV($_)\n";
  }
#   print F "#endif\n";
  print F "\n";
  close(F);
  my $object = $file;
  $object =~ s/\.c/.o/;
  push @obj,$object;
}

open(F,">games.mak") || die "can't write to games.mak\n";
#   print F "ifdef DRIVER_$name\n";
print F "GAMES = \\\n";
foreach (@obj) {
  print F "\t\$(OBJDIR)/games/$_ \\\n";
}
print F "\n";
print F "USE_68K = 1\n" if ($use_68k);
print F "USE_020 = 1\n" if ($use_020);
print F "USE_Z80 = 1\n" if ($use_z80);
print F "USE_6502 = 1\n" if ($use_6502);
print F "USE_2151 = 1\n" if ($use_2151);
print F "USE_2203 = 1\n" if ($use_2203);
print F "USE_2413 = 1\n" if ($use_2413);
print F "USE_2610 = 1\n" if ($use_2610);
print F "USE_3812 = 1\n" if ($use_3812);
print F "USE_ADPCM = 1\n" if ($use_adpcm);
print F "USE_AY8910 = 1\n" if ($use_ay8910);
print F "USE_DAC = 1\n" if ($use_dac);
print F "USE_DXSMP = 1\n" if ($use_dxsmp);
print F "USE_ENSONIQ = 1\n" if ($use_ensoniq);
print F "USE_M6585 = 1\n" if ($use_m6585);
print F "USE_MSM5205 = 1\n" if ($use_msm5205);
print F "USE_NAMCO = 1\n" if ($use_namco);
print F "USE_QSOUND = 1\n" if ($use_qsound);
print F "USE_SMP16BIT = 1\n" if ($use_smp16bit);
print F "USE_YMZ280B = 1\n" if ($use_ymz280b);
#   print F "endif\n";
close(F);
